// ======================================================================
// Simple program to read/analyze time series data for plotting with
// gnuplot etc.

// Copyright (c) 2008-2020 Paul Clark

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// Licensed under GPLv3
// ======================================================================

#define VERSION "0.9.6"

#include <iostream>
#include <vector>
#include <string>
#include <stdlib.h>
#include <math.h>
#include <sstream>
#include <iomanip>
#include <algorithm>

using namespace std;

#define TEST_PATTERN_START 1900
#define TEST_PATTERN_LENGTH 1200
#define PI 3.14159265

//-----------------------------------------------------------------------
// Usage
void usage()
{
    cout << "Time-series data analyzer v" << VERSION << "\n";
    cout << "Copyright (c) 2008-2017 Paul Clark\n";
    cout << "Licenced under GPL v3\n\n";

    cout << "Usage:\n";
    cout << "  analyse [output options] <input format> [step...] < data.txt > output.txt\n\n";
    cout << "Reads data in the given format on stdin:\n";
    cout << "  wti | gnuplot    Basic two-column data (like gnuplot format)\n";
    cout << "  hadcrut3         HADCRUT3 monthly average data (e.g. hadcrut3vgl.txt)\n";
    cout << "  hadcrut4         HADCRUT4 monthly average (e.g. hadcrut4_monthly_ns_avg.txt)\n";
    cout << "  gistemp          GISS surface temperature means (e.g. GLB.Ts+dSST.txt)\n";
    cout << "  uah              UAH NSSTC lower troposphere means (e.g. tltglhmam_5.6)\n";
    cout << "  uah-land         UAH LT land means (e.g. uahncdc.lt, second column)\n";
    cout << "  rss              Remote Sensing Systems MSU monthly zonal averages\n";
    cout << "  best             Berkeley Earth Surface Temperature monthly average\n";
    cout << "  best-upper       BEST 95% upper confidence bound\n";
    cout << "  best-lower       BEST 95% lower confidence bound\n";
    cout << "  sidc-ssn         SIDC month sunspot numbers (e.g. SN_m_tot_V2.0.txt)\n";
    cout << "  esrl             NOAA ESRL Mauna Loa CO2 data (e.g. co2_mm_mlo.txt)\n";
    cout << "  pmod             PMOD daily Solar Irradiance (e.g. composite_d25_07_0310.dat)\n";
    cout << "  pdo              JISAO monthly PDO index (e.g. PDO.latest)\n";
    cout << "  amo              ERSL monthly AMO index (e.g. amon.us.long.data)\n";
    cout << "  nsidc            NSIDC sea ice index (combined, sorted)\n";
    cout << "  covid-19 {country|all}  Covid-19 cases from CSSE\n";
    cout << "  sine {<n>|1}     Sine wave of 1200 samples with n complete cycles(s)\n";
    cout << "  square {<n>|1}   Square wave of 1200 samples with n complete wave(s)\n";
    cout << "  sawtooth {<n>|1} Sawtooth wave of 1200 samples with n complete wave(s)\n";
    cout << "  triangle {<n>|1} Triangle wave of 1200 samples with n complete wave(s)\n";
    cout << "  noise            Pseudo-random noise, 1200 samples\n";

    cout << "\nApplies the following optional processing steps:\n";
    cout << "  from <year>      Cut data before a given year\n";
    cout << "  to <year>        Cut data after a given year\n";
    cout << "  every <n>        Select every n'th sample\n";
    cout << "  last <n>         Select the last n samples\n";
    cout << "  mean <n>         Smooth with n-sample running mean\n";
    cout << "  isolate <n>      Isolate signal apart from n-sample running mean\n";
    cout << "  compress <n>     Average and compress to n-sample units\n";
    cout << "  derivative       Generate first derivative (differences)\n";
    cout << "  integral         Generate integral (cumulative)\n";
    cout << "  scale <n>        Scale samples by n\n";
    cout << "  offset <n>       Offset samples by n\n";
    cout << "  normalise        Normalise all samples to range -0.5..0.5\n";
    cout << "  detrend <n>      Remove a linear trend of 0..n over the range\n";
    cout << "  trend            Derive linear least-squares trend\n";
    cout << "  log              Log (base 10)\n";
    cout << "  fourier          Fourier transform to frequency domain\n";
    cout << "  inverse-fourier  Inverse Fourier transform back to time domain\n";
    cout << "  low-pass <n>     Zero samples from N, symmetrically (use in frequency domain)\n";
    cout << "  high-pass <n>    Zero samples below N, symmetrically (use in frequency domain)\n";
    cout << "  window           Apply Hann window to data\n";
    cout << "  magnitude        Take magnitude of complex Fourier data\n";

    cout << "\nYou can specify multiple steps and they will be run in order\n";

    cout << "\nOutput options:\n";
    cout << "  -a --audio       Output PCM audio (normalised signed words, machine order)\n";
    cout << "  -r --repeat <n>  Repeat data n times\n";
    cout << "\nThe default output is gnuplot format on stdout\n";
}

//-----------------------------------------------------------------------
// Minimal complex number
struct complex
{
  double re;
  double im;
  complex(double _re, double _im=0.0): re(_re), im(_im) {}
  double magnitude() { return sqrt(re*re+im*im); }
};

//-----------------------------------------------------------------------
// Global data
vector<complex> data;       // One value per month, anomaly value
double data_start;         // Decimal year where data starts
double data_interval;      // Interval of data, in years
bool imaginary_valid = false;
double real_data_start;    // Data start from time domain
double real_data_interval; // Data interval from time domain
bool output_comments = true;
bool high_precision_date = false;

//-----------------------------------------------------------------------
// Helper: read a line from cin
// Copes with DOS line endings as well
string read_line()
{
  string s;

  // Read to end of line (variable number of fields with *'s)
  while (cin)
  {
    int c=cin.get();
    if (c=='\n') break;
    if (c>0 && c!='\r') s+=c;
  }

  return s;
}

//-----------------------------------------------------------------------
// Read wti two-column (gnuplot) data
//   - Ignores lines that don't begin with a numeric year
//   - Reads decimal year
//   - Reads one sample
void read_wti()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    string line = read_line();
    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue; // Ignore line without a year at front

    // This is the base year if first
    if (!data_start) data_start = year;

    // Read one sample
    double value=-999;
    iss >> value;
    if (!iss.fail() && value != -999)
      data.push_back(complex(value));
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read HADCRUT3 monthly average data from cin - format:
//   Two lines per year
//   Each line has year at front, followed by 12 values
//   First line of each pair is temperature anomaly
//   Second line is coverage %
void read_hadcrut3()
{
  int expected_year = -1;
  int line = 0;

  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    // Read a line
    int year=-1;
    cin >> year;

    if (cin.fail() || year<0) break;  // end of data

    // This is the base year if first
    if (expected_year < 0)
    {
      expected_year = year;
      data_start = year;
    }

    // Otherwise make sure it's sane
    if (year != expected_year)
    {
      cerr << "Expected year " << expected_year << " but got " << year
	   << " at line " << line+1 << endl;
      exit(2);
    }

    vector<double> monthly;

    // Read 12 months of data
    for(int i=0; i<12 && cin; i++)
    {
      double value=-999;
      cin >> value;
      monthly.push_back(value);
    }

    // Skip to end of line, discarding annual average and whatever the last
    // figure is in CRUTEM4 data!
    while (cin && cin.get() != '\n')
      ;

    // Read and discard repeated year
    cin >> year;

    // Now read 12 month's coverage figures, and post monthly value only
    // if non-zero
    for(int i=0; i<12 && cin; i++)
    {
      double value = 0;
      cin >> value;
      if (value)
	data.push_back(complex(monthly[i]));
      else
	cerr << "Ignored 0 coverage data for " << year+(double)i/12 << endl;
    }

    expected_year++;
    line+=2;
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read HADCRUT4
//   - Ignores lines that don't begin with a numeric year
//   - Reads year/month
//   - Reads first sample (median)
void read_hadcrut4()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    string line = read_line();
    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue; // Ignore line without a year at front

    // Skip the slash
    char c;
    iss >> c;
    if (c!='/') continue;

    double month = -1;
    iss >> month;
    if (iss.fail() || month < 0) continue;

    // This is the base year if first
    if (!data_start) data_start = year+(month-1)/12;

    // Read one sample
    double value=-999;
    iss >> value;
    if (!iss.fail() && value != -999)
      data.push_back(complex(value));
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read SIDC SSN monthly data from cin - format:
//   One line per year, comprising
//     YYYY MM (ignored)
//     Decimal year
//     Sunspot number
//     SD
//     Number of observations
//     (possible *)

// We just read the first decimal year to get the base, and all 
// the raw numbers, ignoring the *s
void read_sidc_ssn()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    // Read and ignore the YYYY MM date
    int date=-1;
    cin >> date >> date;
    if (cin.fail() || date<0) break;  // end of data

    // Get the decimal year
    double year = -1;
    cin >> year;
    if (cin.fail() || year<0) break;

    // This is the base year if first
    if (!data_start) data_start = year;

    // Read sunspot number
    double spots = -1;
    cin >> spots;
    if (cin.fail() || spots < 0) break;

    data.push_back(complex(spots));

    // Read to end of line (variable number of fields with *'s)
    read_line();
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read GISTEMP
//   - Ignores lines that don't begin with a digit
//   - Reads year, then 12 samples (*100)
//   - Ignores rest of line
//   - Stops on *
void read_gistemp()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    string line = read_line();

    // Ignore blank and text lines
    if (line.empty()) continue;
    if (!isdigit(line[0])) continue;

    istringstream iss(line);

    // Get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) break;

    // This is the base year if first
    if (!data_start) data_start = year;

    // Read 12 months of data
    for(int i=0; i<12 && iss; i++)
    {
      double value=-999;
      iss >> value;
      if (!iss.fail() && value != -999)
	data.push_back(complex(value/100));  // Value is in 1/100
      else
	break;  // Stop at *
    }
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read UAH
//   - Ignores lines that don't begin with a numeric year
//   - Reads year and month
//   - Reads first sample (global)
void read_uah()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    string line = read_line();
    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue; // Ignore line without a year at front

    double month = -1;
    iss >> month;
    if (iss.fail() || month < 0) continue;

    // Read one sample
    double value=-999;
    iss >> value;
    if (!iss.fail() && value != -999)
    {
      // This is the base year if first
      if (!data_start) data_start = year+(month-1)/12;

      data.push_back(complex(value));
    }
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read UAH land values
//   - Ignores lines that don't begin with a numeric year
//   - Reads year and month
//   - Reads second sample (land)
void read_uah_land()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    string line = read_line();
    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue; // Ignore line without a year at front

    double month = -1;
    iss >> month;
    if (iss.fail() || month < 0) continue;

    // This is the base year if first
    if (!data_start) data_start = year+(month-1)/12;

    // Skip a sample (global)
    double dummy;
    iss >> dummy;

    // Read one sample
    double value=-999;
    iss >> value;
    if (!iss.fail() && value != -999)
      data.push_back(complex(value));
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read BEST
//   - Ignores lines that begin % or are blank
//   - Reads year and month
//   - Reads first sample (monthly anomaly)
//   - Adds uncertainty * bound
void read_best(int bound=0)
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;
  double last_year = 0;

  while (cin)
  {
    string line = read_line();
    if (line[0] == '%') continue;  // Skip comments

    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue; // Ignore line without a year at front

    // Stop if year rewinds for second dataset
    if (year < last_year) break;
    last_year = year;

    double month = -1;
    iss >> month;
    if (iss.fail() || month < 0) continue;

    // This is the base year if first
    if (!data_start) data_start = year+(month-1)/12;

    // Read one sample
    double value=-999;
    iss >> value;
    if (!iss.fail() && value != -999)
    {
      // Read uncertainty
      double uncertainty = -999;
      iss >> uncertainty;
      if (!iss.fail() && uncertainty != -999) value += bound*uncertainty;
      data.push_back(complex(value));
    }
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read RSS
//   - Ignores lines up to ---
//   - Reads year and month
//   - Reads first sample (global)
void read_rss()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  bool skip = true;
  while (cin)
  {
    string line = read_line();
    // Skip until read -
    if (line[0] == '-')
    {
      skip = false;
      continue;
    }
    else if (skip)
      continue;

    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue; // Ignore line without a year at front

    double month = -1;
    iss >> month;
    if (iss.fail() || month < 0) continue;

    // This is the base year if first
    if (!data_start) data_start = year+(month-1)/12;

    // Read one sample
    double value=-999;
    iss >> value;
    if (!iss.fail() && value != -999)
      data.push_back(complex(value));
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read ESRL CO2
//   - Ignores comment lines with #, and blanks
//   - Skips year and month
//   - Reads decimal year
//   - Skips first sample (average, with gaps)
//   - Reads interpolated data
void read_esrl()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    string line = read_line();

    if (line[0] == '#') continue;

    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue;

    double month = -1;
    iss >> month;
    if (iss.fail() || month < 0) continue;

    // Read year again, as decimal
    iss >> year;

    // This is the base year if first
    if (!data_start) data_start = year;

    // Skip the uncorrected one (which has gaps)
    iss >> year;

    // Read the interpolated value
    double value=-999;
    iss >> value;
    if (!iss.fail() && value != -999)
      data.push_back(complex(value));
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read JISAO-PDO (e.g. PDO.latest)
//   - Ignores lines that don't begin with a digit
//   - Reads year, then up to 12 samples
//   - Ignores *
void read_pdo()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  while (cin)
  {
    string line = read_line();

    // Ignore blank and text lines
    if (line.empty()) continue;
    if (!isdigit(line[0])) continue;

    istringstream iss(line);

    // Get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) break;

    // This is the base year if first
    if (!data_start) data_start = year;

    // Skip any **
    for(;;)
    {
      int c = iss.get();
      if (isspace(c)) break;
    }

    // Read up to 12 months of data
    for(int i=0; i<12 && iss; i++)
    {
      double value=-999;
      iss >> value;
      if (!iss.fail() && value != -999)
	data.push_back(complex(value));
      else
	break;
    }
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read ERSL-AMO (e.g. amon.us.long.data)
//   - Ignores first line
//   - Ignores lines that don't begin with a digit
//   - Reads year, then up to 12 samples
//   - Stops at -99.99
//   - Ignores *
void read_amo()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  read_line();  // Skip start/end year

  while (cin)
  {
    string line = read_line();

    // Ignore blank lines
    if (line.empty()) continue;

    istringstream iss(line);

    // Get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) break;  // Stop at -99.99 missing value line

    // This is the base year if first
    if (!data_start) data_start = year;

    // Read up to 12 months of data
    for(int i=0; i<12 && iss; i++)
    {
      double value=-999;
      iss >> value;
      if (!iss.fail() && value > -99)  // -99.99 is missing value
	data.push_back(complex(value));
      else
	break;
    }
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read PMOD
//   - Ignores first line (comment line count)
//   - Ignores comment lines with ; and blanks
//   - Reads YYMMDD day
//   - Ignores epoch day
//   - Reads TSI value, ignoring -99
//   - Averages all values for a given month
void read_pmod()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  // Skip first line
  read_line();

  double last_t = -1;
  double n_days = 0;
  double sum = 0;

  while (cin)
  {
    string line = read_line();

    if (line[0] == ';') continue;

    istringstream iss(line);

    // Read YYMMDD day
    int day = -1;
    iss >> day;
    if (iss.fail() || day<0) continue;

    // Split into year and month (ignoring day)
    int year = day / 10000;
    int month = (day / 100) % 100;

    // Fix for Y2K
    if (year < 78)
      year += 2000;
    else
      year += 1900;

    // Construct decimal year
    double t = year + (double)(month-1)/12;

    // Check if we've wrapped into another month
    if (t != last_t)
    {
      if (last_t < 0)  // First sample
	real_data_start = data_start = t;
      else
      {
	// Add averaged monthly amount
	if (n_days)
	  data.push_back(complex(sum/n_days));
	else
	  data.push_back(-999);  // No data for this month

	n_days = sum = 0;
      }

      last_t = t;
    }

    // Skip epoch day
    double value;
    iss >> value;

    // Read value
    iss >> value;

    // Accumulate daily value into month
    if (!iss.fail() && value >= 0)
    {
      sum += value;
      n_days++;
    }
  }

  // Add averaged monthly amount of trailing month
  if (n_days)
    data.push_back(complex(sum/n_days));

  cerr << "Read " << data.size() << " monthly averages to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//-----------------------------------------------------------------------
// Read NSIDC
//   - Ignores lines that don't begin with a numeric year
//   - Reads year and month
//   - Reads third field (coverage)
// Interpolates missing months
void read_nsidc()
{
  real_data_start = data_start = 0;
  real_data_interval = data_interval = 1.0/12;

  int n = 0;
  double last_value = 0;

  while (cin)
  {
    string line = read_line();
    istringstream iss(line);

    // Try to get the year
    double year = -1;
    iss >> year;
    if (iss.fail() || year<0) continue; // Ignore line without a year at front

    double month = -1;
    iss >> month;
    if (iss.fail() || month < 0) continue;

    // Skip over two strings
    string skip;
    iss >> skip >> skip;

    // Read coverage data
    double value=-999;
    iss >> value;

    // Skip -9999 as if missing
    if (iss.fail() || value < 0) continue;

    if (data_start)
    {
      // Check for missing data
      int data_n = (int)((year+(month-1)/12-data_start)*12);
      int missing = data_n - n;

      if (missing > 0)
      {
	cerr << "Missing " << missing << " samples before "
	     << year << ":" << month << " - interpolating\n";

	// Interpolate between values
	double slope = (value-last_value)/(missing+1);
	for(int i=1; i<=missing; i++)
	  data.push_back(complex(last_value+slope*i));

	n+=missing;
      }
    }
    else
    {
      // This is the base year if first
      data_start = year+(month-1)/12;
    }

    data.push_back(complex(value));

    n++;
    last_value = value;
  }

  cerr << "Read " << data.size() << " months' values to "
       << data_start+((double)data.size()-1)/12 << endl;
}

//--------------------------------------------------------------------------
// Helper to read a line of CSV into the given strings
// Won't fail, will try to fix up errors
void split_csv(const string& line, vector<string>& vars)
{
  if (line.empty()) return;  // special case to avoid always adding empty var
  string var;
  bool in_quote{false};
  bool escaped{false};
  for(auto c: line)
  {
    if (escaped)
    {
      escaped = false;
      if (c=='"') // doubled quote
      {
        var+=c;
        continue;
      }

      // Not another quote so leave quoting and use normally
      in_quote = false;
    }

    if (c==',' && !in_quote)  // Non-quoted separator
    {
      vars.push_back(var);
      var.clear();
    }
    else if (c=='"')
    {
      if (in_quote)
        escaped = true;  // Need lookahead to see if it is another quote or end
      else
        in_quote = true;
    }
    else var+=c;
  }

  vars.push_back(var);  // Always use remnant, even if empty
}

//-----------------------------------------------------------------------
// Read Covid-19 CSV
//   - Filters for given country
//   - *Daily* values from 22/1/20
//   - Skips region, country, lat, long, reads all from there
void read_covid_19(const string& country)
{
  real_data_start = data_start = 2020+22/366.0;  // Leap year?!
  real_data_interval = data_interval = 1.0/365;  // Average?!
  high_precision_date = true;

  vector<int> totals;

  // Throw away header
  read_line();

  while (cin)
  {
    string line = read_line();
    istringstream iss(line);

    // Split on comma
    vector<string> fields;
    split_csv(line, fields);
    if (fields.size() < 4) continue;

    string cn = fields[1];
    // Replace spaces with - for comparison
    replace(cn.begin(), cn.end(), ' ', '-' );

    if (country == cn || country=="all")
    {
      if (totals.size() < fields.size()-4)
        totals.resize(fields.size()-4);
      for(int i=4; i<fields.size(); i++)
        totals[i-4] += atoi(fields[i].c_str());
    }
  }

  for(int i=0; i<totals.size(); i++)
    data.push_back(complex(totals[i]));

  if (output_comments)
    cout << "#Selected country '" << country << "'\n";
  cerr << "Read " << data.size() << " days' values to "
       << data_start+((double)data.size()-1)*data_interval << endl;
}

//-----------------------------------------------------------------------
// Generate sine test pattern
void generate_sine(double n)
{
  if (n<1.0) n=1.0;  // Default to 1 cycle

  real_data_start = data_start = TEST_PATTERN_START;
  real_data_interval = data_interval = 1.0/12;

  for(int i=0; i<TEST_PATTERN_LENGTH; i++)
  {
    double v = sin(2*PI*i*n/TEST_PATTERN_LENGTH)/2;
    data.push_back(complex(v));
  }
}

//-----------------------------------------------------------------------
// Generate square test pattern
void generate_square(double n)
{
  if (n<1.0) n=1.0;  // Default to 1 cycle

  real_data_start = data_start = TEST_PATTERN_START;
  real_data_interval = data_interval = 1.0/12;

  for(int i=0; i<TEST_PATTERN_LENGTH; i++)
  {
    double v = i*n/TEST_PATTERN_LENGTH;
    v = (v-(int)v >= 0.5)?0.5:-0.5;
    data.push_back(complex(v));
  }
}

//-----------------------------------------------------------------------
// Generate sawtooth test pattern
void generate_sawtooth(double n)
{
  if (n<1.0) n=1.0;  // Default to 1 cycle

  real_data_start = data_start = TEST_PATTERN_START;
  real_data_interval = data_interval = 1.0/12;

  for(int i=0; i<TEST_PATTERN_LENGTH; i++)
  {
    double v = i*n/TEST_PATTERN_LENGTH;
    v -= (int)v+0.5;
    data.push_back(complex(v));
  }
}

//-----------------------------------------------------------------------
// Generate triangle test pattern
void generate_triangle(double n)
{
  if (n<1.0) n=1.0;  // Default to 1 cycle

  real_data_start = data_start = TEST_PATTERN_START;
  real_data_interval = data_interval = 1.0/12;

  for(int i=0; i<TEST_PATTERN_LENGTH; i++)
  {
    double v = i*n/TEST_PATTERN_LENGTH;
    v -= (int)v;
    if (v >= 0.5)
      v = 1.5-2*v;
    else
      v = 2*v-0.5;

    data.push_back(complex(v));
  }
}

//-----------------------------------------------------------------------
// Generate noise
void generate_noise()
{
  real_data_start = data_start = TEST_PATTERN_START;
  real_data_interval = data_interval = 1.0/12;

  srandom(1);  // Make repeatable
  for(int i=0; i<TEST_PATTERN_LENGTH; i++)
  {
    double v = (double)random() / RAND_MAX - 0.5;
    data.push_back(complex(v));
  }
}

//-----------------------------------------------------------------------
// Cut up to a given year
void cut_before(double year)
{
  if (year > data_start)
  {
    int delta = (int)ceil((year-data_start)/data_interval);
    if (delta <= data.size())
    {
      data.erase(data.begin(), data.begin()+delta);
      data_start += delta*data_interval;
    }
    else
    {
      data.clear();
      data_start = year;
    }
  }

  if (output_comments) cout << "#Selected data from " << year << endl;
}

//-----------------------------------------------------------------------
// Cut after a given year
void cut_after(double year)
{
  if (year > data_start)
  {
    int delta = (int)ceil((year-data_start)/data_interval);
    if (delta <= data.size())
      data.erase(data.begin()+delta, data.end());
  }

  if (output_comments) cout << "#Selected data up to " << year << endl;
}

//-----------------------------------------------------------------------
// Select every n'th sample
void select_every(int n)
{
  if (n<1) return;
  vector<complex> selected;  // Easier than trying to delete in place!

  for(int i=0; i<data.size(); i+=n)
    selected.push_back(data[i]);

  data=selected;

  // Change interval 
  data_interval *=n; 

  if (output_comments) cout << "#Selected every " << n << " samples\n";
}

//-----------------------------------------------------------------------
// Select the last N samples
void select_last(int n)
{
  int N = data.size();
  if (n<N)
  {
    data.erase(data.begin(), data.begin()+N-n);
    data_start += (N-n)*data_interval;
  }

  if (output_comments) cout << "#Selected last " << n << " samples\n";
}

//-----------------------------------------------------------------------
// Basic running mean - modifies data in place - real only
void running_mean(int n)
{
  vector<complex> averaged;

  // Average to N samples, around centre
  int shift=n/2;
  for(int i=shift; i<data.size()-shift; i++)
  {
    double total = 0;
    for(int j=0; j<n; j++)
      total += data[i-shift+j].re;

    averaged.push_back(complex(total/n));
  }

  // Copy to data, and shift start year
  data = averaged;
  data_start += (double)shift/12;

  if (output_comments)
    cout << "#Averaged with " << n << "-sample running mean\n";
}

//-----------------------------------------------------------------------
// Isolate remaining signal from running mean - modifies data in place
void isolate(int n)
{
  vector<complex> averaged;

  // Average to N samples, around centre
  int shift=n/2;
  for(int i=shift; i<data.size()-shift; i++)
  {
    double total = 0;
    for(int j=0; j<n; j++)
      total += data[i-shift+j].re;

    averaged.push_back(complex(total/n));
  }

  // Now adjust data with average
  for(int i=0; i<data.size()-shift*2; i++)
    data[i+shift].re -= averaged[i].re;

  // Erase the stuff we haven't adjusted at each end
  data.erase(data.begin(), data.begin()+shift);
  data.erase(data.begin()+data.size()-shift, data.end());

  // Shift start year
  data_start += (double)shift/12;

  if (output_comments)
    cout << "#Isolated from " << n << "-sample running mean\n";
}

//-----------------------------------------------------------------------
// Average and compress to n-month segments
void compress(int n)
{
  vector<complex> averaged;

  // Average to N samples, only including full segments
  for(int i=0; i<n*(data.size()/n); i+=n)
  {
    double total = 0;
    for(int j=0; j<n; j++)
      total += data[i+j].re;

    averaged.push_back(complex(total/n));
  }

  // Copy to data, change interval
  data = averaged;
  data_interval *= n;

  if (output_comments)
    cout << "#Averaged and compressed in " << n << "-sample sections\n";
}

//-----------------------------------------------------------------------
// Generate derivative
void derivative()
{
  // Just operate in place, backwards 
  for(int i=data.size(); i>0; i--)
    data[i].re -= data[i-1].re;

  // Remove the first value and shift
  data.erase(data.begin(), data.begin()+1);
  data_start += data_interval;

  if (output_comments)
    cout << "#First derivative (differences) generated\n";
}

//-----------------------------------------------------------------------
// Generate integral
void integral()
{
  // Just operate in place, forwards
  for(int i=1; i<data.size(); i++)
    data[i].re += data[i-1].re;

  if (output_comments)
    cout << "#Integral (cumulative) generated\n";
}

//-----------------------------------------------------------------------
// Generate integral
void log()
{
  // Just operate in place, forwards
  for(int i=0; i<data.size(); i++)
    data[i].re = (data[i].re > 0.0)?log10(data[i].re):0;

  if (output_comments)
    cout << "#Log (base 10)\n";
}

//-----------------------------------------------------------------------
// Scale
void scale(double n)
{
  // Just operate in place, real only
  for(int i=0; i<data.size(); i++)
    data[i].re *= n;

  if (output_comments)
    cout << "#Scaled by " << n << endl;
}

//-----------------------------------------------------------------------
// Offset
void offset(double n)
{
  // Just operate in place, real only
  for(int i=0; i<data.size(); i++)
    data[i].re += n;

  if (output_comments)
    cout << "#Offset by " << n << endl;
}

//-----------------------------------------------------------------------
// Detrend
void detrend(double n)
{
  int N = data.size()-1;
  if (N<1) return;

  // Just operate in place, real only
  for(int i=0; i<=N; i++)
    data[i].re -= i*n/N;

  if (output_comments)
    cout << "#Linear trend of " << n << " removed\n";
}

//-----------------------------------------------------------------------
// Least squares trend line
void trend()
{
  int n = data.size();
  if (n<=2) return;

  // Accumulate various sums
  double sum_x = 0;
  double sum_y = 0;
  double sum_xy = 0;
  double sum_xx = 0;

  for (int i=0; i<n; i++) 
  {
    double x = i;  // Make it simple to recreate points
    double y = data[i].re;

    sum_x += x;
    sum_y += y;
    sum_xy += x*y;
    sum_xx += x*x;
  }

  // Calculate m,b in y=mx+b;
  double m = (sum_x*sum_y - n*sum_xy)/(sum_x*sum_x - n*sum_xx);
  double b = (sum_y - m*sum_x)/n;

  // Reset data to just two points, one at each end
  data.clear();
  data.push_back(complex(b));
  data.push_back(complex(b+m*n));

  // Change interval to be the whole span
  data_interval = (double)n/12;

  if (output_comments)
    cout << "#Least squares trend line; slope = " << (m*12) << " per year\n";
}

//-----------------------------------------------------------------------
// Normalise
void normalise()
{
  double min = 1e10;
  double max = -1e10;
  int N = data.size();

  // Find min and max
  for(int i=0; i<N; i++)
  {
    double v = data[i].re;
    if (v < min) min = v;
    if (v > max) max = v;
  }

  double range = max-min;
  if (range < 1e-10) return;
  
  // Normalise
  for(int i=0; i<N; i++)
    data[i].re = (data[i].re - min)/range - 0.5;

  if (output_comments)
    cout << "#Normalised to -0.5..0.5\n";
}

//-----------------------------------------------------------------------
// Generate forward fourier transform
// Note:  This code is based on the explanation and source at
//   http://local.wasp.uwa.edu.au/~pbourke/other/dft/
// but tidied up and made C++-ish
//
// It is not efficient (O(N*N)), but we mostly work with small datasets and I
// wanted something simple that I had a chance of understanding!
void fourier(bool forward)
{
  vector<complex> out;

  int N = data.size();

  for(int n=0; n<N; n++)
  {
    complex sum(0,0);

    double d = (forward?-n:n)*2*PI/N;
    for(int k=0; k<N; k++)
    {
      complex& t = data[k];
      double c = cos(d*k);
      double s = sin(d*k);
      sum.re += t.re*c - t.im*s;
      sum.im += t.re*s + t.im*c;
    }

    if (forward)
    {
      sum.re /= N;
      sum.im /= N;
    }

    out.push_back(sum);
  }

  // Copy to data, leave baseline as is, expecting it to come back when
  // inverted
  data = out;

  // Fix up data start and interval
  if (forward)
  {
    imaginary_valid = true;
    // Keep current start/interval to be restored on inverse
    real_data_start = data_start;
    real_data_interval = data_interval;
    
    // Set start interval for harmonics
    data_start = 0;
    data_interval = 1;
  }
  else
  {
    imaginary_valid = false;
    // Restore start/interval to real
    data_start = real_data_start;
    data_interval = real_data_interval;
  }

  if (output_comments)
  {
    if (forward)
      cout << "#Fourier transform to frequency domain\n";
    else
      cout << "#Inverse Fourier transform to time domain\n";
  }
}

//-----------------------------------------------------------------------
// Low pass - zero samples (assumed frequency domain harmonics)
// from n+1 to N-n-1
void low_pass(int n)
{
  for(int i=n+1; i<data.size()-n; i++)
    data[i] = complex(0,0);

  if (output_comments)
    cout << "#Samples zeroed from " << n << " to " 
	 << data.size()-n << " (low pass filter)\n";
}

//-----------------------------------------------------------------------
// High pass - zero samples up to n-1 and from N-n-1 to N
void high_pass(int n)
{
  for(int i=0; i<n; i++)
    data[i] = complex(0,0);
  for(int i=data.size()-n+1; i<data.size(); i++)
    data[i] = complex(0,0);

  if (output_comments)
    cout << "#Samples zeroed below " << n 
	 << " and above " << data.size()-n << " (high pass filter)\n";
}

//-----------------------------------------------------------------------
// Window - apply Hann window to data to remove Fourier end-effects
void window()
{
  int N = data.size();
  if (N<2) return;

  double f = 2*PI/(N-1);  

  for(int i=0; i<data.size(); i++)
    data[i].re *= 0.5*(1- cos(i*f)); // Hann function

  if (output_comments)
    cout << "#Hann window applied\n";
}

//-----------------------------------------------------------------------
// Magnitude - create magnitude of complex data for power spectrum plots
void magnitude()
{
  for(int i=0; i<data.size(); i++)
    data[i] = complex(data[i].magnitude());

  // Block output of imaginary 
  imaginary_valid = false;

  if (output_comments)
    cout << "#Take complex magnitude\n";
}

//-----------------------------------------------------------------------
// Main function
int main(int argc, char **argv)
{
  if (argc < 2)
  {
    usage();
    return 0;
  }

  bool audio = false;
  int repeat = 1;

  int arg;

  // Check for - options
  for(arg=1; arg<argc && argv[arg][0]=='-'; arg++)
  {
    string opt = argv[arg];

    if (opt == "-a" || opt == "--audio")
    {
      audio = true;
      output_comments = false;
    }
    else if ((opt == "-r" || opt == "--repeat") && ++arg<argc)
      repeat = atoi(argv[arg]);
    else
    {
      cerr << "Unknown option: " << opt << endl;
      return 2;
    }
  }

  // Get the format
  string format = argv[arg++];

  // Check for format parameter
  double n=0;

  // Get numeric parameter if given
  if (arg<argc && (isdigit(argv[arg][0]) || argv[arg][0]=='-'))
    n = atof(argv[arg++]);
  
  // Read the input data
  if (format == "wti" || format == "gnuplot")
    read_wti();
  else if (format == "hadcrut3")
    read_hadcrut3();
  else if (format == "hadcrut4")
    read_hadcrut4();
  else if (format == "gistemp")
    read_gistemp();
  else if (format == "uah")
    read_uah();
  else if (format == "uah-land")
    read_uah_land();
  else if (format == "rss")
    read_rss();
  else if (format == "best")
    read_best();
  else if (format == "best-upper")
    read_best(1.0);
  else if (format == "best-lower")
    read_best(-1.0);
  else if (format == "sidc-ssn")
    read_sidc_ssn();
  else if (format == "esrl")
    read_esrl();
  else if (format == "pmod")
    read_pmod();
  else if (format == "pdo")
    read_pdo();
  else if (format == "amo")
    read_amo();
  else if (format == "nsidc")
    read_nsidc();
  else if (format == "covid-19" && arg<argc)
    read_covid_19(argv[arg++]);
  else if (format == "sine")
    generate_sine(n);
  else if (format == "square")
    generate_square(n);
  else if (format == "sawtooth")
    generate_sawtooth(n);
  else if (format == "triangle")
    generate_triangle(n);
  else if (format == "noise")
    generate_noise();
  else
  {
    cerr << "Unknown data format '" << format << "'\n";
    return 2;
  }

  if (high_precision_date)
    cout << setprecision(9);

  // Log comment
  if (output_comments)
    cout << "#Time series (" << format << ") from " << data_start << " to " 
	 << data_start+((double)data.size()-1)*data_interval << endl;

  // Apply steps in order
  while (arg<argc)
  {
    string step = argv[arg++];
    string param;
    double n = 0;

    // Get string and numeric parameter if given
    if (arg<argc && (isdigit(argv[arg][0]) || argv[arg][0]=='-'))
      n = atof(argv[arg++]);

    if (step == "from")
      cut_before(n);
    else if (step == "to")
      cut_after(n);
    else if (step == "every")
      select_every((int)n);
    else if (step == "last")
      select_last((int)n);
    else if (step == "mean")
      running_mean((int)n);
    else if (step == "isolate")
      isolate((int)n);
    else if (step == "compress")
      compress((int)n);
    else if (step == "derivative")
      derivative();
    else if (step == "integral")
      integral();
    else if (step == "scale")
      scale(n);
    else if (step == "offset")
      offset(n);
    else if (step == "normalise")
      normalise();
    else if (step == "detrend")
      detrend(n);
    else if (step == "trend")
      trend();
    else if (step == "log")
      log();
    else if (step == "fourier")
      fourier(true);
    else if (step == "inverse-fourier")
      fourier(false);
    else if (step == "low-pass")
      low_pass((int)n);
    else if (step == "high-pass")
      high_pass((int)n);
    else if (step == "window")
      window();
    else if (step == "magnitude")
      magnitude();
    else
    {
      cerr << "Unknown processing step '" << step << "'\n";
      return 2;
    }
  }

  // Force normalise data for audio
  if (audio) normalise();

  // Output data
  int count = 0;
  double total = 0;
  for(int r=0; r<repeat; r++)
  {
    for(int i=0; i<data.size(); i++)
    {
      if (audio)
      {
	// Output sample as signed words
	short n = (short)(data[i].re*65535);

	// Write in machine order
	cout.write((const char *)&n, 2);
      }
      else
      {
	cout << data_start+i*data_interval << "\t" << data[i].re;
	// Output imaginary if valid
	if (imaginary_valid) cout << "\t" << data[i].im;
	cout << endl;
      }

      count++;
      total += data[i].re;
    }
  }

  if (output_comments)
  {
    cout << "#Data ends\n";
    cout << "#Number of samples: " << count << endl;
    if (count) cout << "#Mean: " << total/count << endl;
  }

  return 0;
}
