// ======================================================================
// Simple program to merge multiple datasets into an overall index

// Copyright (c) 2008 Paul Clark

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// Licensed under GPLv3
// ======================================================================

#define VERSION "0.1.1"

#include <iostream>
#include <vector>
#include <string>
#include <stdlib.h>
#include <math.h>
#include <sstream>

using namespace std;

//-----------------------------------------------------------------------
// Usage
void usage()
{
    cout << "Time-series data merger v" << VERSION << "\n";
    cout << "Copyright (c) 2008 Paul Clark\n";
    cout << "Licenced under GPL v3\n\n";

    cout << "Usage:\n";
    cout << "  merge [process] < data.txt > output.txt\n\n";

    cout << "Data must be in gnuplot format (as output by 'analyse'), separated by 'e'\n";
    cout << "Output is a single gnuplot data series\n\n";

    cout << "Process should be one of:\n";
    cout << "  mean             Take arithmetic mean of series\n";

    cout << "\nData is only included where all series have a value\n";
}

//-----------------------------------------------------------------------
// Series structure
struct Series
{
  vector<double> data;    // One value per month
  double data_start;      // Decimal year where data starts

  Series(double _data_start=0): data_start(_data_start) {}
};

//-----------------------------------------------------------------------
// Global data
vector<Series> inputs;
Series output;

//-----------------------------------------------------------------------
// Helper: read a line from cin
// Copes with DOS line endings as well
string read_line()
{
  string s;

  // Read to end of line (variable number of fields with *'s)
  while (cin)
  {
    int c=cin.get();
    if (c=='\n') break;
    if (c>0 && c!='\r') s+=c;
  }

  return s;
}

//-----------------------------------------------------------------------
// Read multiple gnuplot data series
//   - Ignores lines that don't begin with a numeric year
//   - Reads first year as start point
//   - Splits series on 'e'
void read_input()
{
  bool new_series = true;

  while (cin)
  {
    string line = read_line();
    istringstream iss(line);

    // Skip comments
    if (line[0]=='#') continue;

    // Check for end
    if (line[0]=='e')
    {
      new_series = true;
      continue;
    }

    // Try to get the year
    double year = -1;
    iss >> year;
    if (year<0) continue;  // Ignore anything without a year at front

    // Got a value - create series if required
    if (new_series)
    { 
      inputs.push_back(Series(year));
      new_series = false;
    }

    // Read value
    double value=-999;
    iss >> value;
    if (value != -999) 
      inputs.back().data.push_back(value);  
  }

  cerr << "Read " << inputs.size() << " data series\n";
}

//-----------------------------------------------------------------------
// Synchronise the inputs, removing any values that aren't shared between
// all the series
void synchronise()
{
  // First work out the latest start time 
  double latest_start = 0;

  for(int i=0; i<inputs.size(); i++)
  {
    Series& series = inputs[i];
    if (series.data_start > latest_start) latest_start = series.data_start;
  }

  cerr << "Latest start time is " << latest_start << endl;

  // Now chop out anything before that, and also work out the minimum length
  unsigned min_length = 999999;
  for(int i=0; i<inputs.size(); i++)
  {
    Series& series = inputs[i];
    if (series.data_start < latest_start)
    {
      // Round up
      int delta = (int)((latest_start-series.data_start)*12+0.99);

      if (delta < series.data.size())
	series.data.erase(series.data.begin(), series.data.begin()+delta);
      else
	series.data.clear();

      series.data_start = latest_start;

      // Check min length after chopping off front
      if (series.data.size() < min_length) min_length = series.data.size();
    }
  }

  cerr << "Shortest length is " << min_length << endl;

  // Finally chop off anything over the minium length
  for(int i=0; i<inputs.size(); i++)
  {
    Series& series = inputs[i];
    if (series.data.size() > min_length)
      series.data.erase(series.data.begin()+min_length, series.data.end());
  } 
}

//-----------------------------------------------------------------------
// 'mean' process - take mean of all inputs
void mean()
{
  output.data_start = inputs[0].data_start;

  for(int i=0;i<inputs[0].data.size(); i++)
  {
    double total=0;
    for(int j=0; j<inputs.size(); j++)
      total += inputs[j].data[i];

    output.data.push_back(total/inputs.size());
  }
}

//-----------------------------------------------------------------------
// Main function
int main(int argc, char **argv)
{
  if (argc < 2)
  {
    usage();
    return 0;
  }

  int arg;

  // Check for - options
  for(arg=1; arg<argc && argv[arg][0]=='-'; arg++)
  {
    string opt = argv[arg];

    // Er, there aren't any at the moment
    if (0)
    {
    }
    else
    {
      cerr << "Unknown option: " << opt << endl;
      return 2;
    }
  }

  // Get the process
  string process = argv[arg++];

  // Read input
  read_input();
  if (!inputs.size())
  {
    cerr << "Can't read any data series\n";
    return 4;
  }

  // Synchronise the inputs
  synchronise();

  // Do process
  if (process == "mean")
    mean();

  // Output data
  for(int i=0; i<output.data.size(); i++)
    cout << output.data_start+(i/12.0) << "\t" << output.data[i] << endl;

  return 0;
}
